<?php
require_once dirname(__FILE__) . DIRECTORY_SEPARATOR . 'convoy_response.php';

/**
 * Convoy API
 *
 * @link https://aleksa.tf Aleksa Đorđić
 * @link https://code-cats.com Code Cats
 */
class ConvoyApi
{
    /**
     * @var string Placeholder description
     */
    private $secret;

    /**
     * @var string Placeholder description
     */
    private $url;

    // The data sent with the last request served by this API
    private $lastRequest = [];

    /**
     * Initializes the request parameter
     *
     * @param string $secret Placeholder description
     * @param string $url Placeholder description
     */
    public function __construct($secret, $url)
    {
        $this->secret = $secret;
        $this->url = $url;
    }

    /**
     * Send an API request to Convoy
     *
     * @param string $route The path to the API method
     * @param array $body The data to be sent
     * @param string $method Data transfer method (POST, GET, PUT, DELETE)
     * @return ConvoyResponse
     */
    public function apiRequest($route, array $body, $method)
    {
        $url = $this->url . '/api/application/' . $route;
        $curl = curl_init();

        switch (strtoupper($method)) {
            case 'DELETE':
                // Set data using get parameters
            case 'GET':
                $url .= empty($body) ? '' : '?' . http_build_query($body);
                break;
            case 'POST':
                curl_setopt($curl, CURLOPT_POST, 1);
                // Use the default behavior to set data fields
            default:
                curl_setopt($curl, CURLOPT_POSTFIELDS, json_encode($body));
                break;
        }

        curl_setopt($curl, CURLOPT_CUSTOMREQUEST, strtoupper($method));
        curl_setopt($curl, CURLOPT_RETURNTRANSFER, 1);
        curl_setopt($curl, CURLOPT_HEADER, 1);
        curl_setopt($curl, CURLOPT_URL, $url);
        curl_setopt($curl, CURLOPT_CONNECTTIMEOUT, 30);
        curl_setopt($curl, CURLOPT_SSL_VERIFYPEER, 1);
        curl_setopt($curl, CURLOPT_SSL_VERIFYHOST, 2);
        curl_setopt($curl, CURLOPT_SSLVERSION, 1);

        $headers = [
            'Authorization: Bearer ' . $this->secret,
            'Accept: application/json',
            'Content-Type: application/json'
        ];
        curl_setopt($curl, CURLOPT_HTTPHEADER, $headers);

        $this->lastRequest = ['content' => $body, 'headers' => $headers];
        $result = curl_exec($curl);
        if (curl_errno($curl)) {
            $error = [
                'error' => 'Curl Error',
                'message' => 'An internal error occurred, or the server did not respond to the request.',
                'status' => 500
            ];

            return new ConvoyResponse(['content' => json_encode($error), 'headers' => []]);
        }
        curl_close($curl);

        $data = explode("\n", $result);

        // Return request response
        return new ConvoyResponse([
            'content' => $data[count($data) - 1],
            'headers' => array_splice($data, 0, count($data) - 1)]
        );
    }

    /**
     * Fetch a specific node from Convoy
     * 
     * @param int $id The id of the node to fetch
     * @return ConvoyResponse
     */
    public function getNode($id)
    {
        return $this->apiRequest('nodes/' . $id, [], 'GET');
    }

    /**
     * Fetch nodes from Convoy
     * 
     * @param int $page The page to fetch
     * @return ConvoyResponse
     */
    public function getNodes($page = 1)
    {
        return $this->apiRequest('nodes', ['page' => $page], 'GET');
    }

    /**
     * Fetch all nodes from Convoy
     * 
     * @return array Array of all fetched nodes, null if failed request
     */
    public function getAllNodes()
    {
        $nodes = [];
        $page = 1;

        $counter = 0;
        while ($counter < 1000)
        {
            $request = $this->apiRequest('nodes', ['page' => $page], 'GET');
            if ($request->status() != '200') {
                return null;
            }

            $nodes = array_merge($nodes, $request->response()->data);
            $page++;
            $counter++;

            if ($page > $request->response()->meta->pagination->total_pages) {
                break;
            }
        }
        
        return $nodes;
    }

    /**
     * Creates a user account
     * 
     * @param string $name Account name, max 40 characters
     * @param string $email Account email
     * @param string $password Account password, at least 8 characters
     * @return ConvoyResponse
     */
    public function createUser($name, $email, $password)
    {
        return $this->apiRequest('users', [
            'name' => $name,
            'email' => $email,
            'password' => $password,
            'root_admin' => false
        ], 'POST');
    }

    /**
     * Fetch a server from Convoy
     * 
     * @param int $server_id ID of the server to fetch
     * @return ConvoyResponse
     */
    public function getServer($server_id)
    {
        return $this->apiRequest('servers/' . $server_id, [], 'GET');
    }

    /**
     * Fetch servers from Convoy
     * 
     * @param int $page The page to fetch
     * @return ConvoyResponse
     */
    public function getServers($page = 1)
    {
        return $this->apiRequest('servers', ['page' => $page], 'GET');
    }

    /**
     * Fetch all servers from convoy on specified nodes
     * 
     * @param array $nodes IDs of nodes to fetch servers by
     * @return array Array of all fetched nodes, null if failed request
     */
    public function getServersByNode($nodes)
    {
        $servers = [];
        $page = 1;

        $counter = 0;
        while ($counter < 1000)
        {
            $request = $this->apiRequest('servers', ['page' => $page, 'filter[node_id]' => join(',', $nodes)], 'GET');
            if ($request->status() != '200') {
                return null;
            }

            $servers = array_merge($servers, $request->response()->data);
            $page++;
            $counter++;

            if ($page > $request->response()->meta->pagination->total_pages) {
                break;
            }
        }
        
        return $servers;
    }

    /**
     * Fetch server details
     * 
     * @param int $id Server ID to fetch details of
     * @return ConvoyResponse 
     */
    public function getServerDetails($id)
    {
        return $this->apiRequest('servers/' . $id . '/details', [], 'GET');
    }

    /**
     * Create a server
     * 
     * @param string $name Server name
     * @param string $hostname Server hostname
     * @param string $password Default user password
     * @param int $user_id ID of the Convoy User to add the server to
     * @param int $node_id ID of the Convoy Node to provision the server on
     * @param int $template_uuid UUID of the Template on the Node to provision the server based on
     * @param array $limits Server limits
     *      - cpu
     *      - memory
     *      - disk
     *      - bandwidth
     *      - snapshot
     *      - backup
     * @param array $ipv4_address_ids IDs of the IPv4 addresses to assign to this server
     * @param array $ipv6_address_ids IDs of the IPv6 addresses to assign to this server
     * @return ConvoyResponse
     */
    public function createServer($name, $hostname, $password, $user_id, $node_id, $template_uuid, $limits, $ip_address_ids)
    {
        return $this->apiRequest('servers', [
            'node_id' => $node_id,
            'user_id' => $user_id,
            'name' => $name,
            'hostname' => $hostname,
            'vmid' => null,
            'limits' => [
                'cpu' => $limits['cpu'],
                'memory' => $limits['memory'],
                'disk' => $limits['disk'],
                'bandwidth' => $limits['bandwidth'],
                'snapshots' => $limits['snapshot'],
                'backups' => $limits['backup'],
                'address_ids' => $ip_address_ids
            ],
            'account_password' => $password,
            'template_uuid' => $template_uuid,
            'should_create_server' => true,
            'start_on_completion' => true
        ], 'POST');
    }

    /**
     * Updates a server general information
     * 
     * @param int $server_uuid Server UUID inside of Convoy
     * @param string $name Server name
     * @param string $hostname Server hostname
     */
    public function updateServerInformation($server_uuid, $name, $hostname)
    {
        return $this->apiRequest('servers/' . $server_uuid, [
            'name' => $name,
            'hostname' => $hostname
        ], 'PATCH');
    }

    /**
     * Suspends the specified server
     * 
     * @param int $server_uuid Server UUID inside of Convoy
     * @return ConvoyResponse
     */
    public function suspendServer($server_uuid)
    {
        return $this->apiRequest('servers/' . $server_uuid . '/settings/suspend', [], 'POST');
    }

    /**
     * Unsuspends the specified server
     * 
     * @param int $server_uuid Server UUID inside of Convoy
     * @return ConvoyResponse
     */
    public function unsuspendServer($server_uuid)
    {
        return $this->apiRequest('servers/' . $server_uuid . '/settings/unsuspend', [], 'POST');
    }

    /**
     * Deletes and purges the specified server
     * 
     * @param int $server_uuid Server UUID inside of Convoy
     * @return ConvoyResponse
     */
    public function deleteServer($server_uuid)
    {
        return $this->apiRequest('servers/' . $server_uuid, [
            'no_purge' => false
        ], 'DELETE');
    }

    /**
     * Fetch all templates groups on a node
     * 
     * @param int $node_id ID of the node to select templates groups from
     * @return ConvoyResponse
     */
    public function getTemplates($node_id)
    {
        return $this->apiRequest('nodes/' . $node_id . '/template-groups', [], 'GET');
    }

    /**
     * Fetch the specified address from a node
     * 
     * @param int $node_id ID of the node to select the address from
     * @param int $address_id ID of the address to fetch
     */
    public function getAddress($node_id, $address_id)
    {
        return $this->apiRequest('nodes/' . $node_id . '/addresses/' . $address_id, [], 'GET');
    }

    /**
     * Fetch address by an address from a node
     * 
     * @param int $node_id ID of the node to select the addresses from
     * @param string $address Address to select
     */
    public function getAddressesByAddress($node_id, $address)
    {
        return $this->apiRequest('nodes/' . $node_id . '/addresses', ['filter[address]' => $address], 'GET');
    }

    /**
     * Fetch all addresses from a node
     * 
     * @param int $node_id ID of the node to select addresses from
     * @return ConvoyResponse
     */
    public function getAddresses($node_id)
    {
        return $this->apiRequest('nodes/' . $node_id . '/addresses', [], 'GET');
    }

    /**
     * Creates and fetches an SSO token for the specified user
     * 
     * @param int $user_id User ID of the user to create a login token for
     * @return ConvoyResponse
     */
    public function createSSOToken($user_id)
    {
        return $this->apiRequest('users/' . $user_id . '/generate-sso-token', [], 'POST');
    }
}
